"""
JQ JSON Filter Class
"""

import json  # TODO: Part of the fix for #1059
import os
import re

import pyjq
from  _pyjq import ScriptRuntimeError


class JQRuntimeError(Exception):
    pass



_DEFAULT_LIBRARY_PATH = None

def _library_path():

    global _DEFAULT_LIBRARY_PATH

    if _DEFAULT_LIBRARY_PATH is None:

        _DEFAULT_LIBRARY_PATH = [os.path.expanduser("~/.jq")]

        # All path elements with an executable jq
        origins = list(filter(
            lambda p: os.access(os.path.join(p, "jq"), os.X_OK),
            os.environ["PATH"].split(os.pathsep)
        ))

        # All potential library directories
        libdirs_potential = [
            [
                "%s/%s" % (origin, suffix)
                for suffix in ["../lib/jq", "../lib"]
            ]
            for origin in origins
        ]

        # All library directories that actually exist
        _DEFAULT_LIBRARY_PATH.extend( list(filter(
            lambda d: os.path.isdir(d),
            [item for sublist in libdirs_potential for item in sublist]
        )) )

    return _DEFAULT_LIBRARY_PATH



_import_include = r"((import|include) [^;]+;)"

def _groom(script):
    """
    Groom a filter by moving all imports and includes to the top.  
    """

    # Pull and hold all imports
    lines = [x[0] for x in re.findall(_import_include, script)]
    # Add the rest of the script without the imports and includes
    lines.append(re.sub(_import_include, "", script))
    return "\n".join(lines)



class JQFilter(object):
    """
    JQ JSON filter
    """

    def __init__(
            self,
            filter_spec=".",
            args={},
            output_raw=False,
            groom=False,
            strip_errors_to=None
            ):
        """Construct a filter.  Arguments:

        filter_spec - The JQ script to be used for this filter.  This
        may be any subclass of basestring, a list or a dict.  Strings
        are interpreted directly, lists are stringified and joined
        with newlines (to make multi-line scripts readable in JSON)
        and dicts give their "script" and "output-raw" elements
        extracted and used as if they were either of the other types.

        args - A dictionary of variables to be pre-set in the script.

        output_raw - True to produce raw output instead of JSON.

        groom - Move all 'import' and 'include' statements to the
        top of the script before compiling.  This allows filters
        prepending functions to user-provided scripts that do
        imports to compile properly.

        strip_errors_to - Strip all lines in error messages up to the
        string provided.  This is used to weed known-correct,
        program-provided code out of errors on the first line so the
        user sees the correct line numbers.
        """

        self.output_raw = output_raw

        if isinstance(filter_spec, dict):
            self.output_raw = filter_spec.get("output-raw", output_raw)
            filter_spec = filter_spec.get("script", ".")

        if isinstance(filter_spec, list):
            filter_spec = "\n".join([str(line) for line in filter_spec]) if filter_spec else '.'

        if not isinstance(filter_spec, str):
            raise ValueError("Filter spec must be plain text, list or dict")

        if strip_errors_to is not None and not isinstance(strip_errors_to, str):
            raise ValueError("Strip string must be a string")

        # Passing an args hash into PyJQ causes a memory corruption
        # problem.  As a temporary workaround, turn them into "as"
        # statements.  (See #1059)
        # TODO: Remove this and 'import json' once PyJQ is fixed.

        if len(args):
            arg_lines = "".join([
                "%s as $%s | " % (json.dumps(value), name)
                for name, value in args.items()
            ])
            filter_spec = arg_lines + filter_spec
            args = {}

        if groom:
            filter_spec = _groom(filter_spec)

        value_error = None
        try:
            self.script = pyjq.compile(filter_spec, args, library_paths=_library_path())
        except ValueError as ex:
            # This is held and thrown seprately because Python will
            # produce a confusing nested exception message when it's
            # thrown.
            value_error = ex

        if value_error is not None:

            if strip_errors_to is not None:
                ex_lines = []
                strip_length = len(strip_errors_to)
                for num, line in enumerate(str(value_error).split('\n')):
                    place = line.find(strip_errors_to)
                    if place != -1:
                        line = line[place+strip_length:]
                    ex_lines.append(line.rstrip())
                value_error = ValueError('\n'.join(ex_lines))

            raise value_error



    def __call__(
            self,
            json={}
    ):
        """
        Filter 'json' according to the script.  If output_raw is True,
        join everything that comes out of the filter into a a single
        string and return that.
        """

        try:

            result = self.script.all(json)

            if isinstance(result, list) and self.output_raw:
                return "\n".join([str(item) for item in result])

            elif isinstance(result, dict) or isinstance(result, list):
                return result

            else:
                raise ValueError("No idea what to do with %s result", type(result))

        except ScriptRuntimeError as ex:
            raise JQRuntimeError(str(ex))




if __name__ == "__main__":

    # Check out grooming

    print("Groom Check:")
    print(_groom("def xyz: 123 end; import x/y/z; xyz"))
    print()

    print("Filter:")
    filter = JQFilter(".")
    print(list(filter('{ "foo": 123, "bar": 456 }')))
    print()

    # Note that this only works if pscheduler-jq-library is installed.
    print("Groomed Filter:")
    filter = JQFilter([
        'def x: 123;',
        'import "pscheduler/si" as si;',
        'si::as_integer("12ki"), x'
        ], groom=True)

    print(list(filter('{ "foo": 123, "bar": 456 }')))
    print()
